# Copyright (c) 2018 NEC, Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from oslo_config import cfg
from oslo_upgradecheck import common_checks
from oslo_upgradecheck import upgradecheck

from aodh.i18n import _

CONF = cfg.CONF


class Checks(upgradecheck.UpgradeCommands):

    """Contains upgrade checks

    Various upgrade checks should be added as separate methods in this class
    and added to _upgrade_checks tuple.
    """

    _upgrade_checks = (
        (_('policy File JSON to YAML Migration'),
         (common_checks.check_policy_json, {'conf': CONF})),
    )


def main():
    return upgradecheck.main(
        CONF, project='aodh', upgrade_command=Checks())


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
#
# Copyright 2012 New Dream Network (DreamHost)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Test base classes.
"""
import fixtures
import functools
import os.path
import unittest

import oslo_messaging.conffixture
from oslo_utils import timeutils
from oslotest import base
import webtest

import aodh
from aodh import messaging


class BaseTestCase(base.BaseTestCase):
    def setup_messaging(self, conf, exchange=None):
        self.useFixture(oslo_messaging.conffixture.ConfFixture(conf))
        conf.set_override("notification_driver", ["messaging"])
        if not exchange:
            exchange = 'aodh'
        conf.set_override("control_exchange", exchange)

        # NOTE(sileht): Ensure a new oslo.messaging driver is loaded
        # between each tests
        self.transport = messaging.get_transport(conf, "fake://", cache=False)
        self.useFixture(fixtures.MockPatch(
            'aodh.messaging.get_transport',
            return_value=self.transport))

    def assertTimestampEqual(self, first, second, msg=None):
        """Checks that two timestamps are equals.

        This relies on assertAlmostEqual to avoid rounding problem, and only
        checks up the first microsecond values.

        """
        return self.assertAlmostEqual(
            timeutils.delta_seconds(first, second),
            0.0,
            places=5)

    def assertIsEmpty(self, obj):
        try:
            if len(obj) != 0:
                self.fail("%s is not empty" % type(obj))
        except (TypeError, AttributeError):
            self.fail("%s doesn't have length" % type(obj))

    def assertIsNotEmpty(self, obj):
        try:
            if len(obj) == 0:
                self.fail("%s is empty" % type(obj))
        except (TypeError, AttributeError):
            self.fail("%s doesn't have length" % type(obj))

    def assertDictContains(self, parent, child):
        """Checks whether child dict is a subset of parent.

        assertDictContainsSubset() in standard Python 2.7 has been deprecated
        since Python 3.2
        """
        self.assertEqual(parent, dict(parent, **child))

    @staticmethod
    def path_get(project_file=None):
        root = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                            '..',
                                            '..',
                                            )
                               )
        if project_file:
            return os.path.join(root, project_file)
        else:
            return root

    def assert_single_item(self, items, **filters):
        return self.assert_multiple_items(items, 1, **filters)[0]

    def assert_multiple_items(self, items, count, **filters):
        def _matches(item, **props):
            for prop_name, prop_val in props.items():
                v = (item[prop_name] if isinstance(item, dict)
                     else getattr(item, prop_name))
                if v != prop_val:
                    return False
            return True

        filtered_items = list(
            [item for item in items if _matches(item, **filters)]
        )
        found = len(filtered_items)

        if found != count:
            self.fail("Wrong number of items found [filters=%s, "
                      "expected=%s, found=%s]" % (filters, count, found))

        return filtered_items


def _skip_decorator(func):
    @functools.wraps(func)
    def skip_if_not_implemented(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except aodh.NotImplementedError as e:
            raise unittest.SkipTest(str(e))
        except webtest.app.AppError as e:
            if 'not implemented' in str(e):
                raise unittest.SkipTest(str(e))
            raise
    return skip_if_not_implemented


class SkipNotImplementedMeta(type):
    def __new__(cls, name, bases, local):
        for attr in local:
            value = local[attr]
            if callable(value) and (
                    attr.startswith('test_') or attr == 'setUp'):
                local[attr] = _skip_decorator(value)
        return type.__new__(cls, name, bases, local)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""SQLAlchemy storage backend."""
import copy
import datetime
import os.path

from alembic import command
from alembic import config
from alembic import migration
from oslo_db.sqlalchemy import enginefacade
from oslo_db.sqlalchemy import utils as oslo_sql_utils
from oslo_log import log
from oslo_utils import importutils
from oslo_utils import timeutils
import sqlalchemy
from sqlalchemy import asc
from sqlalchemy import desc
from sqlalchemy.engine import url as sqlalchemy_url
from sqlalchemy import func
from sqlalchemy.orm import exc

import aodh
from aodh import storage
from aodh.storage import base
from aodh.storage import models as alarm_api_models
from aodh.storage.sqlalchemy import models
from aodh.storage.sqlalchemy import utils as sql_utils


osprofiler_sqlalchemy = importutils.try_import('osprofiler.sqlalchemy')

LOG = log.getLogger(__name__)

AVAILABLE_CAPABILITIES = {
    'alarms': {'query': {'simple': True,
                         'complex': True},
               'history': {'query': {'simple': True,
                                     'complex': True}}},
}


AVAILABLE_STORAGE_CAPABILITIES = {
    'storage': {'production_ready': True},
}


def apply_filters(query, model, **filters):
    filter_dict = {}

    for key, value in filters.items():
        column_attr = getattr(model, key)

        if isinstance(value, dict):
            if 'in' in value:
                query = query.filter(column_attr.in_(value['in']))
            elif 'nin' in value:
                query = query.filter(~column_attr.in_(value['nin']))
            elif 'ne' in value:
                query = query.filter(column_attr != value['ne'])
            elif 'gt' in value:
                query = query.filter(column_attr > value['gt'])
            elif 'ge' in value:
                query = query.filter(column_attr >= value['ge'])
            elif 'lt' in value:
                query = query.filter(column_attr < value['lt'])
            elif 'le' in value:
                query = query.filter(column_attr <= value['le'])
            elif 'eq' in value:
                query = query.filter(column_attr == value['eq'])
            elif 'has' in value:
                like_pattern = '%{0}%'.format(value['has'])
                query = query.filter(column_attr.like(like_pattern))
        else:
            filter_dict[key] = value

    if filter_dict:
        query = query.filter_by(**filter_dict)

    return query


class Connection(base.Connection):
    """Put the data into a SQLAlchemy database. """
    CAPABILITIES = base.update_nested(base.Connection.CAPABILITIES,
                                      AVAILABLE_CAPABILITIES)
    STORAGE_CAPABILITIES = base.update_nested(
        base.Connection.STORAGE_CAPABILITIES,
        AVAILABLE_STORAGE_CAPABILITIES,
    )

    def __init__(self, conf, url):
        # Set max_retries to 0, since oslo.db in certain cases may attempt
        # to retry making the db connection retried max_retries ^ 2 times
        # in failure case and db reconnection has already been implemented
        # in storage.__init__.get_connection_from_config function
        options = dict(conf.database.items())
        options['max_retries'] = 0
        # FIXME(stephenfin): Remove this (and ideally use of
        # LegacyEngineFacade) asap since it's not compatible with SQLAlchemy
        # 2.0
        options['autocommit'] = True
        # oslo.db doesn't support options defined by Aodh
        for opt in storage.OPTS:
            options.pop(opt.name, None)
        self._engine_facade = enginefacade.LegacyEngineFacade(
            self.dress_url(url),
            **options)

        if osprofiler_sqlalchemy:
            osprofiler_sqlalchemy.add_tracing(sqlalchemy,
                                              self._engine_facade.get_engine(),
                                              'db')
        self.conf = conf

    @staticmethod
    def dress_url(url):
        # If no explicit driver has been set, we default to pymysql
        if url.startswith("mysql://"):
            url = sqlalchemy_url.make_url(url)
            url = url.set(drivername="mysql+pymysql")
            return str(url)
        return url

    def disconnect(self):
        self._engine_facade.get_engine().dispose()

    def _get_alembic_config(self):
        cfg = config.Config(
            "%s/sqlalchemy/alembic/alembic.ini" % os.path.dirname(__file__))
        cfg.set_main_option('sqlalchemy.url',
                            self.conf.database.connection.replace("%", "%%"))
        return cfg

    def upgrade(self, nocreate=False):
        cfg = self._get_alembic_config()
        cfg.conf = self.conf
        if nocreate:
            command.upgrade(cfg, "head")
        else:
            engine = self._engine_facade.get_engine()
            ctxt = migration.MigrationContext.configure(engine.connect())
            current_version = ctxt.get_current_revision()
            if current_version is None:
                models.Base.metadata.create_all(engine, checkfirst=False)
                command.stamp(cfg, "head")
            else:
                command.upgrade(cfg, "head")

    def clear(self):
        engine = self._engine_facade.get_engine()
        for table in reversed(models.Base.metadata.sorted_tables):
            engine.execute(table.delete())
        engine.dispose()

    def _retrieve_data(self, filter_expr, orderby, limit, table):
        if limit == 0:
            return []

        session = self._engine_facade.get_session()
        engine = self._engine_facade.get_engine()
        query = session.query(table)
        transformer = sql_utils.QueryTransformer(table, query,
                                                 dialect=engine.dialect.name)
        if filter_expr is not None:
            transformer.apply_filter(filter_expr)

        transformer.apply_options(orderby,
                                  limit)

        retrieve = {models.Alarm: self._retrieve_alarms,
                    models.AlarmChange: self._retrieve_alarm_history}
        return retrieve[table](transformer.get_query())

    @staticmethod
    def _row_to_alarm_model(row):
        return alarm_api_models.Alarm(
            alarm_id=row.alarm_id,
            enabled=row.enabled,
            type=row.type,
            name=row.name,
            description=row.description,
            timestamp=row.timestamp,
            user_id=row.user_id,
            project_id=row.project_id,
            state=row.state,
            state_timestamp=row.state_timestamp,
            state_reason=row.state_reason,
            ok_actions=row.ok_actions,
            alarm_actions=row.alarm_actions,
            insufficient_data_actions=(row.insufficient_data_actions),
            rule=row.rule,
            time_constraints=row.time_constraints,
            repeat_actions=row.repeat_actions,
            severity=row.severity,
            evaluate_timestamp=row.evaluate_timestamp
        )

    def _retrieve_alarms(self, query):
        return [self._row_to_alarm_model(x) for x in query.all()]

    @staticmethod
    def _get_pagination_query(session, query, pagination, api_model, model):
        if not pagination.get('sort'):
            pagination['sort'] = api_model.DEFAULT_SORT
        marker = None
        if pagination.get('marker'):
            key_attr = getattr(model, api_model.PRIMARY_KEY)
            marker_query = copy.copy(query)
            marker_query = marker_query.filter(
                key_attr == pagination['marker'])
            try:
                marker = marker_query.limit(1).one()
            except exc.NoResultFound:
                raise storage.InvalidMarker(
                    'Marker %s not found.' % pagination['marker'])
        limit = pagination.get('limit')
        # we sort by "severity" by its semantic than its alphabetical
        # order when "severity" specified in sorts.
        for sort_key, sort_dir in pagination['sort'][::-1]:
            if sort_key == 'severity':
                engine = session.connection()
                if engine.dialect.name != "mysql":
                    raise aodh.NotImplementedError
                sort_dir_func = {'asc': asc, 'desc': desc}[sort_dir]
                query = query.order_by(sort_dir_func(
                    func.field(getattr(model, sort_key), 'low',
                               'moderate', 'critical')))
                pagination['sort'].remove((sort_key, sort_dir))

        sort_keys = [s[0] for s in pagination['sort']]
        sort_dirs = [s[1] for s in pagination['sort']]
        return oslo_sql_utils.paginate_query(
            query, model, limit, sort_keys, sort_dirs=sort_dirs, marker=marker)

    def get_alarms(self, meter=None, pagination=None, **kwargs):
        """Yields a lists of alarms that match filters."""
        pagination = pagination or {}
        session = self._engine_facade.get_session()
        query = session.query(models.Alarm)
        query = apply_filters(query, models.Alarm, **kwargs)
        query = self._get_pagination_query(
            session, query, pagination, alarm_api_models.Alarm, models.Alarm)

        alarms = self._retrieve_alarms(query)

        # TODO(cmart): improve this by using sqlalchemy.func factory
        if meter is not None:
            alarms = filter(lambda row:
                            row.rule.get('meter_name', None) == meter,
                            alarms)

        return alarms

    def create_alarm(self, alarm):
        """Create an alarm.

        :param alarm: The alarm to create.
        """
        session = self._engine_facade.get_session()
        with session.begin():
            alarm_row = models.Alarm(alarm_id=alarm.alarm_id)
            alarm_row.update(alarm.as_dict())
            session.add(alarm_row)

        return self._row_to_alarm_model(alarm_row)

    def update_alarm(self, alarm):
        """Update an alarm.

        :param alarm: the new Alarm to update
        """
        session = self._engine_facade.get_session()
        with session.begin():
            count = session.query(models.Alarm).filter(
                models.Alarm.alarm_id == alarm.alarm_id).update(
                    alarm.as_dict())
            if not count:
                raise storage.AlarmNotFound(alarm.alarm_id)
        return alarm

    def delete_alarm(self, alarm_id):
        """Delete an alarm and its history data.

        :param alarm_id: ID of the alarm to delete
        """
        session = self._engine_facade.get_session()
        with session.begin():
            session.query(models.Alarm).filter(
                models.Alarm.alarm_id == alarm_id).delete()
            # FIXME(liusheng): we should use delete cascade
            session.query(models.AlarmChange).filter(
                models.AlarmChange.alarm_id == alarm_id).delete()

    @staticmethod
    def _row_to_alarm_change_model(row):
        return alarm_api_models.AlarmChange(event_id=row.event_id,
                                            alarm_id=row.alarm_id,
                                            type=row.type,
                                            detail=row.detail,
                                            user_id=row.user_id,
                                            project_id=row.project_id,
                                            on_behalf_of=row.on_behalf_of,
                                            timestamp=row.timestamp,
                                            severity=row.severity)

    def query_alarms(self, filter_expr=None, orderby=None, limit=None):
        """Yields a lists of alarms that match filter."""
        return self._retrieve_data(filter_expr, orderby, limit, models.Alarm)

    def _retrieve_alarm_history(self, query):
        return (self._row_to_alarm_change_model(x) for x in query.all())

    def query_alarm_history(self, filter_expr=None, orderby=None, limit=None):
        """Return an iterable of model.AlarmChange objects."""
        return self._retrieve_data(filter_expr,
                                   orderby,
                                   limit,
                                   models.AlarmChange)

    def get_alarm_changes(self, alarm_id, on_behalf_of,
                          user=None, project=None, alarm_type=None,
                          severity=None, start_timestamp=None,
                          start_timestamp_op=None, end_timestamp=None,
                          end_timestamp_op=None, pagination=None):
        """Yields list of AlarmChanges describing alarm history

        Changes are always sorted in reverse order of occurrence, given
        the importance of currency.

        Segregation for non-administrative users is done on the basis
        of the on_behalf_of parameter. This allows such users to have
        visibility on both the changes initiated by themselves directly
        (generally creation, rule changes, or deletion) and also on those
        changes initiated on their behalf by the alarming service (state
        transitions after alarm thresholds are crossed).

        :param alarm_id: ID of alarm to return changes for
        :param on_behalf_of: ID of tenant to scope changes query (None for
                             administrative user, indicating all projects)
        :param user: Optional ID of user to return changes for
        :param project: Optional ID of project to return changes for
        :param alarm_type: Optional change type
        :param severity: Optional alarm severity
        :param start_timestamp: Optional modified timestamp start range
        :param start_timestamp_op: Optional timestamp start range operation
        :param end_timestamp: Optional modified timestamp end range
        :param end_timestamp_op: Optional timestamp end range operation
        :param pagination: Pagination query parameters.
        """
        pagination = pagination or {}
        session = self._engine_facade.get_session()
        query = session.query(models.AlarmChange)
        query = query.filter(models.AlarmChange.alarm_id == alarm_id)

        if on_behalf_of is not None:
            query = query.filter(
                models.AlarmChange.on_behalf_of == on_behalf_of)
        if user is not None:
            query = query.filter(models.AlarmChange.user_id == user)
        if project is not None:
            query = query.filter(models.AlarmChange.project_id == project)
        if alarm_type is not None:
            query = query.filter(models.AlarmChange.type == alarm_type)
        if severity is not None:
            query = query.filter(models.AlarmChange.severity == severity)
        if start_timestamp:
            if start_timestamp_op == 'gt':
                query = query.filter(
                    models.AlarmChange.timestamp > start_timestamp)
            else:
                query = query.filter(
                    models.AlarmChange.timestamp >= start_timestamp)
        if end_timestamp:
            if end_timestamp_op == 'le':
                query = query.filter(
                    models.AlarmChange.timestamp <= end_timestamp)
            else:
                query = query.filter(
                    models.AlarmChange.timestamp < end_timestamp)

        query = self._get_pagination_query(
            session, query, pagination, alarm_api_models.AlarmChange,
            models.AlarmChange)
        return self._retrieve_alarm_history(query)

    def record_alarm_change(self, alarm_change):
        """Record alarm change event."""
        session = self._engine_facade.get_session()
        with session.begin():
            alarm_change_row = models.AlarmChange(
                event_id=alarm_change['event_id'])
            alarm_change_row.update(alarm_change)
            session.add(alarm_change_row)

    def clear_expired_alarm_history_data(self, ttl, max_count=100):
        """Clear expired alarm history data from the backend storage system.

        Clearing occurs according to the time-to-live.

        :param ttl: Number of seconds to keep alarm history records for.
        :param max_count: Number of records to delete.
        """
        session = self._engine_facade.get_session()
        with session.begin():
            end = timeutils.utcnow() - datetime.timedelta(seconds=ttl)
            alarm_history_q = (session.query(models.AlarmChange.event_id)
                               .filter(models.AlarmChange.timestamp < end))
            event_ids = [i[0] for i in alarm_history_q.limit(max_count)]
            deleted_rows = session.query(models.AlarmChange).filter(
                models.AlarmChange.event_id.in_(event_ids)
            ).delete(synchronize_session="fetch")
            LOG.info("%d alarm histories are removed from database",
                     deleted_rows)

    def conditional_update(self, model, values, expected_values, filters=None):
        """Compare-and-swap conditional update SQLAlchemy implementation."""
        filters = filters or {}
        filters.update(expected_values)

        session = self._engine_facade.get_session()
        query = session.query(model)
        if filters:
            query = query.filter_by(**filters)

        update_args = {'synchronize_session': False}

        result = query.update(values, **update_args)
        return 0 != result

    @staticmethod
    def _row_to_quota_model(row):
        return alarm_api_models.Quota(
            project_id=row.project_id,
            resource=row.resource,
            limit=row.limit,
        )

    def _retrieve_quotas(self, query):
        return [self._row_to_quota_model(x) for x in query.all()]

    def get_quotas(self, project_id):
        """Get resource quota for the given project."""
        filters = {'project_id': project_id}
        session = self._engine_facade.get_session()
        query = session.query(models.Quota).filter_by(**filters)
        return self._retrieve_quotas(query)

    def set_quotas(self, project_id, quotas):
        """Set resource quota for the given user."""
        session = self._engine_facade.get_session()

        with session.begin():
            for q in quotas:
                values = {
                    'project_id': project_id,
                    'resource': q['resource'],
                }

                quota = session.query(models.Quota).filter_by(**values).first()
                if not quota:
                    new_quota = models.Quota(project_id=project_id,
                                             resource=q['resource'],
                                             limit=q['limit'])
                    session.add(new_quota)
                else:
                    values['limit'] = q['limit']
                    quota.update(values.copy())

        filters = {'project_id': project_id}
        query = session.query(models.Quota).filter_by(**filters)
        return self._retrieve_quotas(query)

    def delete_quotas(self, project_id):
        filters = {'project_id': project_id}
        session = self._engine_facade.get_session()
        session.query(models.Quota).filter_by(**filters).delete()



# Copyright 2020 Catalyst Cloud LTD.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import copy

from oslo_utils import uuidutils

from aodh.tests.functional.api import v2


class TestQuotas(v2.FunctionalTest):
    @classmethod
    def setUpClass(cls):
        super(TestQuotas, cls).setUpClass()

        cls.project = uuidutils.generate_uuid()
        cls.user = uuidutils.generate_uuid()
        cls.auth_headers = {'X-User-Id': cls.user, 'X-Project-Id': cls.project}
        cls.other_project = uuidutils.generate_uuid()

    def test_get_quotas_by_user(self):
        resp = self.get_json('/quotas', headers=self.auth_headers, status=200)

        self.assertEqual(self.project, resp.get('project_id'))
        self.assertGreater(len(resp.get('quotas', [])), 0)

    def test_get_project_quotas_by_user(self):
        resp = self.get_json('/quotas?project_id=%s' % self.project,
                             headers=self.auth_headers, status=200)

        self.assertEqual(self.project, resp.get('project_id'))
        self.assertGreater(len(resp.get('quotas', [])), 0)

    def test_get_other_project_quotas_by_user_failed(self):
        self.get_json(
            '/quotas?project_id=%s' % self.other_project,
            headers=self.auth_headers,
            expect_errors=True,
            status=401
        )

    def test_get_project_quotas_by_admin(self):
        auth_headers = copy.copy(self.auth_headers)
        auth_headers['X-Roles'] = 'admin'

        resp = self.get_json('/quotas?project_id=%s' % self.other_project,
                             headers=auth_headers,
                             status=200)

        self.assertEqual(self.other_project, resp.get('project_id'))
        self.assertGreater(len(resp.get('quotas', [])), 0)

    def test_post_quotas_by_admin(self):
        auth_headers = copy.copy(self.auth_headers)
        auth_headers['X-Roles'] = 'admin'

        resp = self.post_json(
            '/quotas',
            {
                "project_id": self.other_project,
                "quotas": [
                    {
                        "resource": "alarms",
                        "limit": 30
                    }
                ]
            },
            headers=auth_headers,
            status=201
        )
        resp_json = resp.json

        self.assertEqual(self.other_project, resp_json.get('project_id'))
        self.assert_single_item(resp_json.get('quotas', []), resource='alarms',
                                limit=30)

    def test_post_quotas_by_user_failed(self):
        self.post_json(
            '/quotas',
            {
                "project_id": self.other_project,
                "quotas": [
                    {
                        "resource": "alarms",
                        "limit": 20
                    }
                ]
            },
            headers=self.auth_headers,
            expect_errors=True,
            status=403
        )

    def test_post_quotas_no_limit_failed(self):
        auth_headers = copy.copy(self.auth_headers)
        auth_headers['X-Roles'] = 'admin'

        resp = self.post_json(
            '/quotas',
            {
                "project_id": self.project,
                "quotas": [
                    {
                        "resource": "alarms"
                    }
                ]
            },
            headers=auth_headers,
            expect_errors=True,
            status=400
        )

        self.assertIn('Mandatory field missing',
                      resp.json['error_message']['faultstring'])

    def test_post_quotas_no_resource_failed(self):
        auth_headers = copy.copy(self.auth_headers)
        auth_headers['X-Roles'] = 'admin'

        resp = self.post_json(
            '/quotas',
            {
                "project_id": self.project,
                "quotas": [
                    {
                        "limit": 1
                    }
                ]
            },
            headers=auth_headers,
            expect_errors=True,
            status=400
        )

        self.assertIn('Mandatory field missing',
                      resp.json['error_message']['faultstring'])

    def test_post_quotas_wrong_limit_failed(self):
        auth_headers = copy.copy(self.auth_headers)
        auth_headers['X-Roles'] = 'admin'

        resp = self.post_json(
            '/quotas',
            {
                "project_id": self.project,
                "quotas": [
                    {
                        "resource": "alarms",
                        "limit": -5
                    }
                ]
            },
            headers=auth_headers,
            expect_errors=True,
            status=400
        )

        self.assertIn('Value should be greater or equal to -1',
                      resp.json['error_message']['faultstring'])

    def test_post_quotas_unsupported_resource_failed(self):
        auth_headers = copy.copy(self.auth_headers)
        auth_headers['X-Roles'] = 'admin'

        resp = self.post_json(
            '/quotas',
            {
                "project_id": self.project,
                "quotas": [
                    {
                        "resource": "other_resource",
                        "limit": 1
                    }
                ]
            },
            headers=auth_headers,
            expect_errors=True,
            status=400
        )

        self.assertIn('Value should be one of',
                      resp.json['error_message']['faultstring'])

    def test_delete_project_quota_by_admin(self):
        auth_headers = copy.copy(self.auth_headers)
        auth_headers['X-Roles'] = 'admin'

        self.post_json(
            '/quotas',
            {
                "project_id": self.other_project,
                "quotas": [
                    {
                        "resource": "alarms",
                        "limit": 30
                    }
                ]
            },
            headers=auth_headers,
            status=201
        )

        resp = self.get_json('/quotas?project_id=%s' % self.other_project,
                             headers=auth_headers,
                             status=200)
        self.assert_single_item(resp['quotas'], resource='alarms',
                                limit=30)

        self.delete('/quotas/%s' % self.other_project, headers=auth_headers,
                    status=204)

        resp = self.get_json('/quotas?project_id=%s' % self.other_project,
                             headers=auth_headers,
                             status=200)
        self.assert_multiple_items(resp['quotas'], 0, resource='alarms',
                                   limit=30)

    def test_delete_project_quota_by_user_failed(self):
        self.delete('/quotas/%s' % self.other_project,
                    headers=self.auth_headers,
                    expect_errors=True,
                    status=403)



#
# Copyright 2012 New Dream Network, LLC (DreamHost)
# Copyright 2015-2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import uuid

from oslo_log import log
from paste import deploy
import pecan

from aodh.api import hooks
from aodh.api import middleware
from aodh import service
from aodh import storage

LOG = log.getLogger(__name__)


# NOTE(sileht): pastedeploy uses ConfigParser to handle
# global_conf, since python 3 ConfigParser doesn't
# allow storing object as config value, only strings are
# permit, so to be able to pass an object created before paste load
# the app, we store them into a global var. But the each loaded app
# store it's configuration in unique key to be concurrency safe.
global APPCONFIGS
APPCONFIGS = {}


def setup_app(root, conf):
    app_hooks = [hooks.ConfigHook(conf),
                 hooks.DBHook(
                     storage.get_connection_from_config(conf)),
                 hooks.TranslationHook()]
    return pecan.make_app(
        root,
        hooks=app_hooks,
        wrap_app=middleware.ParsableErrorMiddleware,
        guess_content_type_from_ext=False
    )


def load_app(conf):
    global APPCONFIGS

    # Build the WSGI app
    cfg_path = conf.api.paste_config
    if not os.path.isabs(cfg_path):
        cfg_path = conf.find_file(cfg_path)

    if cfg_path is None or not os.path.exists(cfg_path):
        LOG.debug("No api-paste configuration file found! Using default.")
        cfg_path = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__), "api-paste.ini"))

    config = dict(conf=conf)
    configkey = str(uuid.uuid4())
    APPCONFIGS[configkey] = config

    LOG.info("WSGI config used: %s", cfg_path)
    return deploy.loadapp("config:" + cfg_path,
                          name="aodh+" + (
                              conf.api.auth_mode
                              if conf.api.auth_mode else "noauth"
                          ),
                          global_conf={'configkey': configkey})


def app_factory(global_config, **local_conf):
    global APPCONFIGS
    appconfig = APPCONFIGS.get(global_config.get('configkey'))
    return setup_app(root=local_conf.get('root'), **appconfig)


def build_wsgi_app(argv=None):
    return load_app(service.prepare_service(argv=argv))



# Copyright 2014-2015 eNovance
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import itertools

from keystoneauth1 import loading

import aodh.api
import aodh.api.controllers.v2.alarm_rules.gnocchi
import aodh.api.controllers.v2.alarms
import aodh.coordination
import aodh.evaluator
import aodh.evaluator.event
import aodh.evaluator.gnocchi
import aodh.evaluator.loadbalancer
import aodh.evaluator.threshold
import aodh.event
import aodh.keystone_client
import aodh.notifier.rest
import aodh.notifier.zaqar
import aodh.service
import aodh.storage


def list_opts():
    return [
        ('DEFAULT',
         itertools.chain(
             aodh.evaluator.OPTS,
             aodh.evaluator.event.OPTS,
             aodh.evaluator.threshold.OPTS,
             aodh.evaluator.loadbalancer.OPTS,
             aodh.notifier.rest.OPTS,
             aodh.queue.OPTS,
             aodh.service.OPTS)),
        ('api',
         itertools.chain(
             aodh.api.OPTS,
             aodh.api.controllers.v2.alarm_rules.gnocchi.GNOCCHI_OPTS,
             aodh.api.controllers.v2.alarms.ALARM_API_OPTS)),
        ('coordination', aodh.coordination.OPTS),
        ('database', aodh.storage.OPTS),
        ('evaluator', aodh.service.EVALUATOR_OPTS),
        ('listener', itertools.chain(aodh.service.LISTENER_OPTS,
                                     aodh.event.OPTS)),
        ('notifier', itertools.chain(aodh.notifier.OPTS,
                                     aodh.service.NOTIFIER_OPTS)),
        ('service_credentials', aodh.keystone_client.OPTS),
        ('service_types', aodh.notifier.zaqar.SERVICE_OPTS),
    ]


def list_keystoneauth_opts():
    # NOTE(sileht): the configuration file contains only the options
    # for the password plugin that handles keystone v2 and v3 API
    # with discovery. But other options are possible.
    return [('service_credentials', (
            loading.get_auth_common_conf_options() +
            loading.get_auth_plugin_conf_options('password')))]



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import json

from stevedore import named
from wsme.rest import json as wjson
from wsme import types as wtypes

from aodh.api.controllers.v2 import base
from aodh.i18n import _


class InvalidCompositeRule(base.ClientSideError):
    def __init__(self, error):
        err = _('Invalid input composite rule: %s, it should '
                'be a dict with an "and" or "or" as key, and the '
                'value of dict should be a list of basic threshold '
                'rules or sub composite rules, can be nested.') % error
        super(InvalidCompositeRule, self).__init__(err)


class CompositeRule(wtypes.UserType):
    """Composite alarm rule.

    A simple dict type to preset composite rule.
    """

    basetype = wtypes.text
    name = 'composite_rule'

    threshold_plugins = None

    def __init__(self):
        threshold_rules = ('gnocchi_resources_threshold',
                           'gnocchi_aggregation_by_metrics_threshold',
                           'gnocchi_aggregation_by_resources_threshold')
        CompositeRule.threshold_plugins = named.NamedExtensionManager(
            "aodh.alarm.rule", threshold_rules)
        super(CompositeRule, self).__init__()

    @staticmethod
    def valid_composite_rule(rules):
        if isinstance(rules, dict) and len(rules) == 1:
            and_or_key = list(rules)[0]
            if and_or_key not in ('and', 'or'):
                raise base.ClientSideError(
                    _('Threshold rules should be combined with "and" or "or"'))
            if isinstance(rules[and_or_key], list):
                for sub_rule in rules[and_or_key]:
                    CompositeRule.valid_composite_rule(sub_rule)
            else:
                raise InvalidCompositeRule(rules)
        elif isinstance(rules, dict):
            rule_type = rules.pop('type', None)
            if not rule_type:
                raise base.ClientSideError(_('type must be set in every rule'))

            if rule_type not in CompositeRule.threshold_plugins:
                plugins = sorted(CompositeRule.threshold_plugins.names())
                err = _('Unsupported sub-rule type :%(rule)s in composite '
                        'rule, should be one of: %(plugins)s') % {
                            'rule': rule_type,
                            'plugins': plugins}
                raise base.ClientSideError(err)
            plugin = CompositeRule.threshold_plugins[rule_type].plugin
            wjson.fromjson(plugin, rules)
            rule_dict = plugin(**rules).as_dict()
            rules.update(rule_dict)
            rules.update(type=rule_type)
        else:
            raise InvalidCompositeRule(rules)

    @staticmethod
    def validate(value):
        try:
            json.dumps(value)
        except TypeError:
            raise base.ClientSideError(_('%s is not JSON serializable')
                                       % value)
        else:
            CompositeRule.valid_composite_rule(value)
            return value

    @staticmethod
    def frombasetype(value):
        return CompositeRule.validate(value)

    @staticmethod
    def create_hook(alarm):
        pass

    @staticmethod
    def validate_alarm(alarm):
        pass

    @staticmethod
    def update_hook(alarm):
        pass

    @staticmethod
    def as_dict():
        pass

    @staticmethod
    def __call__(**rule):
        return rule


composite_rule = CompositeRule()



#
# Copyright 2014 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from oslo_config import fixture as fixture_config
import tooz.coordination

from aodh import coordination
from aodh import service
from aodh.tests import base


class MockToozCoordinator(object):
    def __init__(self, member_id, shared_storage):
        self._member_id = member_id
        self._groups = shared_storage
        self.is_started = False

    def start(self):
        self.is_started = True

    def stop(self):
        pass

    def heartbeat(self):
        pass

    def create_group(self, group_id):
        if group_id in self._groups:
            return MockAsyncError(
                tooz.coordination.GroupAlreadyExist(group_id))
        self._groups[group_id] = {}
        return MockAsyncResult(None)

    def join_group(self, group_id, capabilities=b''):
        if group_id not in self._groups:
            return MockAsyncError(
                tooz.coordination.GroupNotCreated(group_id))
        if self._member_id in self._groups[group_id]:
            return MockAsyncError(
                tooz.coordination.MemberAlreadyExist(group_id,
                                                     self._member_id))
        self._groups[group_id][self._member_id] = {
            "capabilities": capabilities,
        }
        return MockAsyncResult(None)

    def leave_group(self, group_id):
        return MockAsyncResult(None)

    def get_members(self, group_id):
        if group_id not in self._groups:
            return MockAsyncError(
                tooz.coordination.GroupNotCreated(group_id))
        return MockAsyncResult(self._groups[group_id])


class MockToozCoordExceptionRaiser(MockToozCoordinator):
    def start(self):
        raise tooz.coordination.ToozError('error')

    def heartbeat(self):
        raise tooz.coordination.ToozError('error')

    def join_group(self, group_id, capabilities=b''):
        raise tooz.coordination.ToozError('error')

    def get_members(self, group_id):
        raise tooz.coordination.ToozError('error')


class MockAsyncResult(tooz.coordination.CoordAsyncResult):
    def __init__(self, result):
        self.result = result

    def get(self, timeout=0):
        return self.result

    @staticmethod
    def done():
        return True


class MockAsyncError(tooz.coordination.CoordAsyncResult):
    def __init__(self, error):
        self.error = error

    def get(self, timeout=0):
        raise self.error

    @staticmethod
    def done():
        return True


class TestHashRing(base.BaseTestCase):
    def test_hash_ring(self):
        num_nodes = 10
        num_keys = 1000

        nodes = [str(x) for x in range(num_nodes)]
        hr = coordination.HashRing(nodes)

        buckets = [0] * num_nodes
        assignments = [-1] * num_keys
        for k in range(num_keys):
            n = int(hr.get_node(str(k)))
            self.assertLessEqual(0, n)
            self.assertLessEqual(n, num_nodes)
            buckets[n] += 1
            assignments[k] = n

        # at least something in each bucket
        self.assertTrue(all((c > 0 for c in buckets)))

        # approximately even distribution
        diff = max(buckets) - min(buckets)
        self.assertLess(diff, 0.3 * (num_keys / num_nodes))

        # consistency
        num_nodes += 1
        nodes.append(str(num_nodes + 1))
        hr = coordination.HashRing(nodes)
        for k in range(num_keys):
            n = int(hr.get_node(str(k)))
            assignments[k] -= n
        reassigned = len([c for c in assignments if c != 0])
        self.assertLess(reassigned, num_keys / num_nodes)


class TestPartitioning(base.BaseTestCase):

    def setUp(self):
        super(TestPartitioning, self).setUp()
        conf = service.prepare_service(argv=[], config_files=[])
        self.CONF = self.useFixture(fixture_config.Config(conf)).conf
        self.shared_storage = {}

    def _get_new_started_coordinator(self, shared_storage, agent_id=None,
                                     coordinator_cls=None):
        coordinator_cls = coordinator_cls or MockToozCoordinator
        self.CONF.set_override('backend_url', 'xxx://yyy',
                               group='coordination')
        with mock.patch('tooz.coordination.get_coordinator',
                        lambda _, member_id:
                        coordinator_cls(member_id, shared_storage)):
            pc = coordination.PartitionCoordinator(self.CONF, agent_id)
            pc.start()
            return pc

    def _usage_simulation(self, *agents_kwargs):
        partition_coordinators = []
        for kwargs in agents_kwargs:
            partition_coordinator = self._get_new_started_coordinator(
                self.shared_storage, kwargs['agent_id'], kwargs.get(
                    'coordinator_cls'))
            partition_coordinator.join_group(kwargs['group_id'])
            partition_coordinators.append(partition_coordinator)

        for i, kwargs in enumerate(agents_kwargs):
            all_resources = kwargs.get('all_resources', [])
            expected_resources = kwargs.get('expected_resources', [])
            actual_resources = partition_coordinators[i].extract_my_subset(
                kwargs['group_id'], all_resources)
            self.assertEqual(expected_resources, actual_resources)

    def test_single_group(self):
        agents = [dict(agent_id='agent1', group_id='group'),
                  dict(agent_id='agent2', group_id='group')]
        self._usage_simulation(*agents)

        self.assertEqual(['group'], sorted(self.shared_storage.keys()))
        self.assertEqual(['agent1', 'agent2'],
                         sorted(self.shared_storage['group'].keys()))

    def test_multiple_groups(self):
        agents = [dict(agent_id='agent1', group_id='group1'),
                  dict(agent_id='agent2', group_id='group2')]
        self._usage_simulation(*agents)

        self.assertEqual(['group1', 'group2'],
                         sorted(self.shared_storage.keys()))

    def test_member_id(self):
        agent = 'agent'.encode('ascii')
        coord = self._get_new_started_coordinator({}, agent)
        self.assertEqual(agent, coord._my_id)

    def test_partitioning(self):
        all_resources = ['resource_%s' % i for i in range(1000)]
        agents = ['agent_%s' % i for i in range(10)]

        expected_resources = [list() for _ in range(len(agents))]
        hr = coordination.HashRing(agents)
        for r in all_resources:
            key = agents.index(hr.get_node(r))
            expected_resources[key].append(r)

        agents_kwargs = []
        for i, agent in enumerate(agents):
            agents_kwargs.append(dict(agent_id=agent,
                                 group_id='group',
                                 all_resources=all_resources,
                                 expected_resources=expected_resources[i]))
        self._usage_simulation(*agents_kwargs)

    @mock.patch.object(coordination.LOG, 'exception')
    def test_coordination_backend_offline(self, mocked_exception):
        agents = [dict(agent_id='agent1',
                       group_id='group',
                       all_resources=['res1', 'res2'],
                       expected_resources=[],
                       coordinator_cls=MockToozCoordExceptionRaiser)]
        self._usage_simulation(*agents)
        called = [mock.call('Error connecting to coordination backend.'),
                  mock.call('Error getting group membership info from '
                            'coordination backend.')]
        self.assertEqual(called, mocked_exception.call_args_list)

    @mock.patch.object(coordination.LOG, 'exception')
    @mock.patch.object(coordination.LOG, 'info')
    def test_reconnect(self, mock_info, mocked_exception):
        coord = self._get_new_started_coordinator({}, 'a',
                                                  MockToozCoordExceptionRaiser)
        with mock.patch('tooz.coordination.get_coordinator',
                        return_value=MockToozCoordExceptionRaiser('a', {})):
            coord.heartbeat()
        called = [mock.call('Error connecting to coordination backend.'),
                  mock.call('Error connecting to coordination backend.'),
                  mock.call('Error sending a heartbeat to coordination '
                            'backend.')]
        self.assertEqual(called, mocked_exception.call_args_list)
        with mock.patch('tooz.coordination.get_coordinator',
                        return_value=MockToozCoordinator('a', {})):
            coord.heartbeat()
        mock_info.assert_called_with('Coordination backend started '
                                     'successfully.')

    def test_group_id_none(self):
        coord = self._get_new_started_coordinator({}, 'a')
        self.assertTrue(coord._coordinator.is_started)

        with mock.patch.object(coord._coordinator, 'join_group') as mocked:
            coord.join_group(None)
            self.assertEqual(0, mocked.call_count)
        with mock.patch.object(coord._coordinator, 'leave_group') as mocked:
            coord.leave_group(None)
            self.assertEqual(0, mocked.call_count)

    def test_stop(self):
        coord = self._get_new_started_coordinator({}, 'a')
        self.assertTrue(coord._coordinator.is_started)
        coord.join_group("123")
        coord.stop()
        self.assertIsEmpty(coord._groups)
        self.assertIsNone(coord._coordinator)



#
# Copyright Ericsson AB 2013. All rights reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Test the methods related to complex query."""
import datetime
from unittest import mock

import fixtures
import jsonschema
from oslotest import base
import wsme

from aodh.api.controllers.v2 import query
from aodh.storage import models as alarm_models


class FakeComplexQuery(query.ValidatedComplexQuery):
    def __init__(self, db_model, additional_name_mapping=None, metadata=False):
        super(FakeComplexQuery, self).__init__(query=None,
                                               db_model=db_model,
                                               additional_name_mapping=(
                                                   additional_name_mapping or
                                                   {}),
                                               metadata_allowed=metadata)


class TestComplexQuery(base.BaseTestCase):
    def setUp(self):
        super(TestComplexQuery, self).setUp()
        self.useFixture(fixtures.MonkeyPatch(
            'pecan.response', mock.MagicMock()))
        self.query = FakeComplexQuery(alarm_models.Alarm)
        self.query_alarmchange = FakeComplexQuery(
            alarm_models.AlarmChange)

    def test_replace_isotime_utc(self):
        filter_expr = {"=": {"timestamp": "2013-12-05T19:38:29Z"}}
        self.query._replace_isotime_with_datetime(filter_expr)
        self.assertEqual(datetime.datetime(2013, 12, 5, 19, 38, 29),
                         filter_expr["="]["timestamp"])

    def test_replace_isotime_timezone_removed(self):
        filter_expr = {"=": {"timestamp": "2013-12-05T20:38:29+01:00"}}
        self.query._replace_isotime_with_datetime(filter_expr)
        self.assertEqual(datetime.datetime(2013, 12, 5, 20, 38, 29),
                         filter_expr["="]["timestamp"])

    def test_replace_isotime_wrong_syntax(self):
        filter_expr = {"=": {"timestamp": "not a valid isotime string"}}
        self.assertRaises(wsme.exc.ClientSideError,
                          self.query._replace_isotime_with_datetime,
                          filter_expr)

    def test_replace_isotime_in_complex_filter(self):
        filter_expr = {"and": [{"=": {"timestamp": "2013-12-05T19:38:29Z"}},
                               {"=": {"timestamp": "2013-12-06T19:38:29Z"}}]}
        self.query._replace_isotime_with_datetime(filter_expr)
        self.assertEqual(datetime.datetime(2013, 12, 5, 19, 38, 29),
                         filter_expr["and"][0]["="]["timestamp"])
        self.assertEqual(datetime.datetime(2013, 12, 6, 19, 38, 29),
                         filter_expr["and"][1]["="]["timestamp"])

    def test_replace_isotime_in_complex_filter_with_unbalanced_tree(self):
        subfilter = {"and": [{"=": {"project_id": 42}},
                             {"=": {"timestamp": "2013-12-06T19:38:29Z"}}]}

        filter_expr = {"or": [{"=": {"timestamp": "2013-12-05T19:38:29Z"}},
                              subfilter]}

        self.query._replace_isotime_with_datetime(filter_expr)
        self.assertEqual(datetime.datetime(2013, 12, 5, 19, 38, 29),
                         filter_expr["or"][0]["="]["timestamp"])
        self.assertEqual(datetime.datetime(2013, 12, 6, 19, 38, 29),
                         filter_expr["or"][1]["and"][1]["="]["timestamp"])

    def test_convert_operator_to_lower_case(self):
        filter_expr = {"AND": [{"=": {"project_id": 42}},
                               {"=": {"project_id": 44}}]}
        self.query._convert_operator_to_lower_case(filter_expr)
        self.assertEqual("and", list(filter_expr.keys())[0])

        filter_expr = {"Or": [{"=": {"project_id": 43}},
                              {"anD": [{"=": {"project_id": 44}},
                                       {"=": {"project_id": 42}}]}]}
        self.query._convert_operator_to_lower_case(filter_expr)
        self.assertEqual("or", list(filter_expr.keys())[0])
        self.assertEqual("and", list(filter_expr["or"][1].keys())[0])

    def test_invalid_filter_misstyped_field_name_samples(self):
        filter = {"=": {"project_id11": 42}}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

    def test_invalid_filter_misstyped_field_name_alarms(self):
        filter = {"=": {"enabbled": True}}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

    def test_invalid_filter_misstyped_field_name_alarmchange(self):
        filter = {"=": {"tpe": "rule change"}}
        self.assertRaises(jsonschema.ValidationError,
                          self.query_alarmchange._validate_filter,
                          filter)

    def test_invalid_complex_filter_wrong_field_names(self):
        filter = {"and":
                  [{"=": {"non_existing_field": 42}},
                   {"=": {"project_id": 42}}]}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

        filter = {"and":
                  [{"=": {"project_id": 42}},
                   {"=": {"non_existing_field": 42}}]}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

        filter = {"and":
                  [{"=": {"project_id11": 42}},
                   {"=": {"project_id": 42}}]}
        self.assertRaises(jsonschema.ValidationError,
                          self.query_alarmchange._validate_filter,
                          filter)

        filter = {"or":
                  [{"=": {"non_existing_field": 42}},
                   {"and":
                    [{"=": {"project_id": 44}},
                     {"=": {"project_id": 42}}]}]}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

        filter = {"or":
                  [{"=": {"project_id": 43}},
                   {"and":
                    [{"=": {"project_id": 44}},
                     {"=": {"non_existing_field": 42}}]}]}
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_filter,
                          filter)

    def test_convert_orderby(self):
        orderby = []
        self.query._convert_orderby_to_lower_case(orderby)
        self.assertEqual([], orderby)

        orderby = [{"project_id": "DESC"}]
        self.query._convert_orderby_to_lower_case(orderby)
        self.assertEqual([{"project_id": "desc"}], orderby)

        orderby = [{"project_id": "ASC"}, {"resource_id": "DESC"}]
        self.query._convert_orderby_to_lower_case(orderby)
        self.assertEqual([{"project_id": "asc"}, {"resource_id": "desc"}],
                         orderby)

    def test_validate_orderby_empty_direction(self):
        orderby = [{"project_id": ""}]
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_orderby,
                          orderby)
        orderby = [{"project_id": "asc"}, {"resource_id": ""}]
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_orderby,
                          orderby)

    def test_validate_orderby_wrong_order_string(self):
        orderby = [{"project_id": "not a valid order"}]
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_orderby,
                          orderby)

    def test_validate_orderby_wrong_multiple_item_order_string(self):
        orderby = [{"project_id": "not a valid order"}, {"resource_id": "ASC"}]
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_orderby,
                          orderby)

    def test_validate_orderby_empty_field_name(self):
        orderby = [{"": "ASC"}]
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_orderby,
                          orderby)
        orderby = [{"project_id": "asc"}, {"": "desc"}]
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_orderby,
                          orderby)

    def test_validate_orderby_wrong_field_name(self):
        orderby = [{"project_id11": "ASC"}]
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_orderby,
                          orderby)

    def test_validate_orderby_wrong_field_name_multiple_item_orderby(self):
        orderby = [{"project_id": "asc"}, {"resource_id11": "ASC"}]
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_orderby,
                          orderby)

    def test_validate_orderby_metadata_is_not_allowed(self):
        orderby = [{"metadata.display_name": "asc"}]
        self.assertRaises(jsonschema.ValidationError,
                          self.query._validate_orderby,
                          orderby)



#
# Copyright 2013 eNovance <licensing@enovance.com>
# Copyright 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

from oslo_config import fixture
from oslotest import base

from aodh import service


class TestEvaluatorBase(base.BaseTestCase):
    def setUp(self):
        super(TestEvaluatorBase, self).setUp()
        conf = service.prepare_service(argv=[], config_files=[])
        self.conf = self.useFixture(fixture.Config(conf)).conf
        self.evaluator = self.EVALUATOR(self.conf)
        self.notifier = mock.MagicMock()
        self.evaluator.notifier = self.notifier
        self.storage_conn = mock.MagicMock()
        self.evaluator.storage_conn = self.storage_conn
        self.evaluator._ks_client = mock.Mock(user_id='fake_user_id',
                                              project_id='fake_project_id',
                                              auth_token='fake_token')
        self.prepare_alarms()

    def prepare_alarms(self):
        self.alarms = []

    def _evaluate_all_alarms(self):
        for alarm in self.alarms:
            self.evaluator.evaluate(alarm)

    def _set_all_alarms(self, state):
        for alarm in self.alarms:
            alarm.state = state

    def _assert_all_alarms(self, state):
        for alarm in self.alarms:
            self.assertEqual(state, alarm.state)

    def assertDictContains(self, parent, child):
        """Checks whether child dict is a subset of parent."""
        self.assertEqual(parent, dict(parent, **child))



#
# Copyright 2015 NEC Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import voluptuous
import wsme
from wsme import types as wtypes

from aodh.api.controllers.v2 import base
from aodh.i18n import _


# Schema validation for the event type query.
_q_validator = voluptuous.Schema(
    {"field": voluptuous.Match(r"^[a-zA-Z.',0-9_-]*$"),
     "op": voluptuous.In(base.operation_kind),
     "value": voluptuous.In(["string", "integer", "float", "boolean", ""])})


class AlarmEventRule(base.AlarmRule):
    """Alarm Event Rule.

    Describe when to trigger the alarm based on an event
    """

    event_type = wsme.wsattr(wtypes.text)
    "The type of event (default is '*')"

    query = wsme.wsattr([base.Query])
    "The query to find the event (default is [])"

    def __init__(self, event_type=None, query=None):
        event_type = event_type or '*'
        query = [base.Query(**q) for q in query or []]
        super(AlarmEventRule, self).__init__(event_type=event_type,
                                             query=query)

    @classmethod
    def validate_alarm(cls, alarm):
        super(AlarmEventRule, cls).validate_alarm(alarm)
        for i in alarm.event_rule.query:
            i.get_value()
            try:
                _q_validator({"field": i.field, "op": i.op,
                              "value": i.type})
            except voluptuous.MultipleInvalid as e:
                raise base.ClientSideError(
                    _("Query value or traits invalid: %s") % str(e))

    @property
    def default_description(self):
        return _('Alarm when %s event occurred.') % self.event_type

    def as_dict(self):
        rule = self.as_dict_from_keys(['event_type'])
        rule['query'] = [q.as_dict() for q in self.query]
        return rule

    @classmethod
    def sample(cls):
        return cls(event_type='compute.instance.update',
                   query=[{'field': 'traits.instance_id"',
                           'value': '153462d0-a9b8-4b5b-8175-9e4b05e9b856',
                           'op': 'eq',
                           'type': 'string'}])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslotest import base

from aodh.api.controllers.v2 import capabilities


class TestCapabilities(base.BaseTestCase):

    def test_recursive_keypairs(self):
        data = {'a': 'A', 'b': 'B',
                'nested': {'a': 'A', 'b': 'B'}}
        pairs = list(capabilities._recursive_keypairs(data))
        self.assertEqual([('a', 'A'), ('b', 'B'),
                          ('nested:a', 'A'), ('nested:b', 'B')],
                         pairs)

    def test_recursive_keypairs_with_separator(self):
        data = {'a': 'A',
                'b': 'B',
                'nested': {'a': 'A',
                           'b': 'B',
                           },
                }
        separator = '.'
        pairs = list(capabilities._recursive_keypairs(data, separator))
        self.assertEqual([('a', 'A'),
                          ('b', 'B'),
                          ('nested.a', 'A'),
                          ('nested.b', 'B')],
                         pairs)

    def test_recursive_keypairs_with_list_of_dict(self):
        small = 1
        big = 1 << 64
        expected = [('a', 'A'),
                    ('b', 'B'),
                    ('nested:list', [{small: 99, big: 42}])]
        data = {'a': 'A',
                'b': 'B',
                'nested': {'list': [{small: 99, big: 42}]}}
        pairs = list(capabilities._recursive_keypairs(data))
        self.assertEqual(len(expected), len(pairs))
        for k, v in pairs:
            # the keys 1 and 1<<64 cause a hash collision on 64bit platforms
            if k == 'nested:list':
                self.assertIn(v,
                              [[{small: 99, big: 42}],
                               [{big: 42, small: 99}]])
            else:
                self.assertIn((k, v), expected)



#
# Copyright 2015 eNovance
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import json

from gnocchiclient import client
from gnocchiclient import exceptions
from oslo_log import log

from aodh.evaluator import threshold
from aodh import keystone_client

LOG = log.getLogger(__name__)

# The list of points that Gnocchi API returned is composed
# of tuples with (timestamp, granularity, value)
GRANULARITY = 1
VALUE = 2


class GnocchiBase(threshold.ThresholdEvaluator):
    def __init__(self, conf):
        super(GnocchiBase, self).__init__(conf)
        self._gnocchi_client = client.Client(
            '1', keystone_client.get_session(conf),
            adapter_options={
                'interface': conf.service_credentials.interface,
                'region_name': conf.service_credentials.region_name})

    @staticmethod
    def _sanitize(rule, statistics):
        """Return the datapoints that correspond to the alarm granularity"""
        # TODO(sileht): if there's no direct match, but there is an archive
        # policy with granularity that's an even divisor or the period,
        # we could potentially do a mean-of-means (or max-of-maxes or whatever,
        # but not a stddev-of-stddevs).
        # TODO(sileht): support alarm['exclude_outliers']
        LOG.debug('sanitize stats %s', statistics)
        # NOTE(jamespage)
        # Dynamic Aggregates are returned in a dict struct so
        # check for this first.
        if isinstance(statistics, dict):
            # Pop array of measures from aggregated subdict
            statistics = statistics['measures']['aggregated']
        statistics = [stats[VALUE] for stats in statistics
                      if stats[GRANULARITY] == rule['granularity']]
        if not statistics:
            raise threshold.InsufficientDataError(
                "No datapoint for granularity %s" % rule['granularity'], [])
        statistics = statistics[-rule['evaluation_periods']:]
        LOG.debug('pruned statistics to %d', len(statistics))
        return statistics


class GnocchiResourceThresholdEvaluator(GnocchiBase):
    def _statistics(self, rule, start, end):
        try:
            return self._gnocchi_client.metric.get_measures(
                metric=rule['metric'],
                granularity=rule['granularity'],
                start=start, stop=end,
                resource_id=rule['resource_id'],
                aggregation=rule['aggregation_method'])
        except exceptions.MetricNotFound:
            raise threshold.InsufficientDataError(
                'metric %s for resource %s does not exists' %
                (rule['metric'], rule['resource_id']), [])
        except exceptions.ResourceNotFound:
            raise threshold.InsufficientDataError(
                'resource %s does not exists' % rule['resource_id'], [])
        except exceptions.NotFound:
            # TODO(sileht): gnocchiclient should raise a explicit
            # exception for AggregationNotFound, this API endpoint
            # can only raise 3 different 404, so we are safe to
            # assume this is an AggregationNotFound for now.
            raise threshold.InsufficientDataError(
                'aggregation %s does not exist for '
                'metric %s of resource %s' % (rule['aggregation_method'],
                                              rule['metric'],
                                              rule['resource_id']),
                [])
        except Exception as e:
            msg = 'alarm statistics retrieval failed: %s' % e
            LOG.warning(msg)
            raise threshold.InsufficientDataError(msg, [])


class GnocchiAggregationMetricsThresholdEvaluator(GnocchiBase):
    def _statistics(self, rule, start, end):
        try:
            _operations = [
                'aggregate', rule['aggregation_method']
            ]
            for metric in rule['metrics']:
                _operations.append(
                    [
                        'metric', metric,
                        rule['aggregation_method'].lstrip('rate:')
                    ]
                )
            # FIXME(sileht): In case of a heat autoscaling stack decide to
            # delete an instance, the gnocchi metrics associated to this
            # instance will be no more updated and when the alarm will ask
            # for the aggregation, gnocchi will raise a 'No overlap'
            # exception.
            # So temporary set 'needed_overlap' to 0 to disable the
            # gnocchi checks about missing points. For more detail see:
            #   https://bugs.launchpad.net/gnocchi/+bug/1479429
            return self._gnocchi_client.aggregates.fetch(
                operations=_operations,
                granularity=rule['granularity'],
                start=start, stop=end,
                needed_overlap=0)
        except exceptions.MetricNotFound:
            raise threshold.InsufficientDataError(
                'At least of metrics in %s does not exist' %
                rule['metrics'], [])
        except exceptions.NotFound:
            # TODO(sileht): gnocchiclient should raise a explicit
            # exception for AggregationNotFound, this API endpoint
            # can only raise 3 different 404, so we are safe to
            # assume this is an AggregationNotFound for now.
            raise threshold.InsufficientDataError(
                'aggregation %s does not exist for at least one '
                'metrics in %s' % (rule['aggregation_method'],
                                   rule['metrics']), [])
        except Exception as e:
            msg = 'alarm statistics retrieval failed: %s' % e
            LOG.warning(msg)
            raise threshold.InsufficientDataError(msg, [])


class GnocchiAggregationResourcesThresholdEvaluator(GnocchiBase):
    def _statistics(self, rule, start, end):
        try:
            # FIXME(sileht): In case of a heat autoscaling stack decide to
            # delete an instance, the gnocchi metrics associated to this
            # instance will be no more updated and when the alarm will ask
            # for the aggregation, gnocchi will raise a 'No overlap'
            # exception.
            # So temporary set 'needed_overlap' to 0 to disable the
            # gnocchi checks about missing points. For more detail see:
            #   https://bugs.launchpad.net/gnocchi/+bug/1479429
            return self._gnocchi_client.aggregates.fetch(
                operations=[
                    'aggregate', rule['aggregation_method'],
                    [
                        'metric', rule['metric'],
                        rule['aggregation_method'].lstrip('rate:')
                    ]
                ],
                granularity=rule['granularity'],
                search=json.loads(rule['query']),
                resource_type=rule["resource_type"],
                start=start, stop=end,
                needed_overlap=0)
        except exceptions.MetricNotFound:
            raise threshold.InsufficientDataError(
                'metric %s does not exists' % rule['metric'], [])
        except exceptions.NotFound:
            # TODO(sileht): gnocchiclient should raise a explicit
            # exception for AggregationNotFound, this API endpoint
            # can only raise 3 different 404, so we are safe to
            # assume this is an AggregationNotFound for now.
            raise threshold.InsufficientDataError(
                'aggregation %s does not exist for at least one '
                'metric of the query' % rule['aggregation_method'], [])
        except Exception as e:
            msg = 'alarm statistics retrieval failed: %s' % e
            LOG.warning(msg)
            raise threshold.InsufficientDataError(msg, [])



#!/usr/bin/env python
#
# Copyright 2012-2015 eNovance <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import subprocess

from oslo_utils import fileutils

from aodh.tests import base


class BinTestCase(base.BaseTestCase):
    def setUp(self):
        super(BinTestCase, self).setUp()
        content = ("[database]\n"
                   "connection=log://localhost\n")
        content = content.encode('utf-8')
        self.tempfile = fileutils.write_to_tempfile(content=content,
                                                    prefix='aodh',
                                                    suffix='.conf')

    def tearDown(self):
        super(BinTestCase, self).tearDown()
        os.remove(self.tempfile)

    def test_dbsync_run(self):
        subp = subprocess.Popen(['aodh-dbsync',
                                 "--config-file=%s" % self.tempfile])
        self.assertEqual(0, subp.wait())

    def test_run_expirer_ttl_disabled(self):
        subp = subprocess.Popen(['aodh-expirer',
                                 '-d',
                                 "--config-file=%s" % self.tempfile],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        out, __ = subp.communicate()
        self.assertEqual(0, subp.poll())
        self.assertIn(b"Nothing to clean, database alarm history "
                      b"time to live is disabled", out)

    def test_run_expirer_ttl_enabled(self):
        content = ("[database]\n"
                   "alarm_history_time_to_live=1\n"
                   "alarm_histories_delete_batch_size=10\n"
                   "connection=log://localhost\n")
        content = content.encode('utf-8')
        self.tempfile = fileutils.write_to_tempfile(content=content,
                                                    prefix='aodh',
                                                    suffix='.conf')
        subp = subprocess.Popen(['aodh-expirer',
                                 '-d',
                                 "--config-file=%s" % self.tempfile],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        out, __ = subp.communicate()
        self.assertEqual(0, subp.poll())
        msg = "Dropping alarm history 10 data with TTL 1"
        msg = msg.encode('utf-8')
        self.assertIn(msg, out)


class BinEvaluatorTestCase(base.BaseTestCase):
    def setUp(self):
        super(BinEvaluatorTestCase, self).setUp()
        content = ("[database]\n"
                   "connection=log://localhost\n")
        content = content.encode('utf-8')
        self.tempfile = fileutils.write_to_tempfile(content=content,
                                                    prefix='aodh',
                                                    suffix='.conf')
        self.subp = None

    def tearDown(self):
        super(BinEvaluatorTestCase, self).tearDown()
        if self.subp:
            self.subp.kill()
        os.remove(self.tempfile)

    def test_starting_evaluator(self):
        self.subp = subprocess.Popen(['aodh-evaluator',
                                      "--config-file=%s" % self.tempfile],
                                     stderr=subprocess.PIPE)
        self.assertIsNone(self.subp.poll())


class BinNotifierTestCase(BinEvaluatorTestCase):
    def test_starting_notifier(self):
        self.subp = subprocess.Popen(['aodh-notifier',
                                      "--config-file=%s" % self.tempfile],
                                     stderr=subprocess.PIPE)
        self.assertIsNone(self.subp.poll())



#
# Copyright 2015 NEC Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import datetime
import json
from unittest import mock

from oslo_utils import timeutils
from oslo_utils import uuidutils

from aodh import evaluator
from aodh.evaluator import event as event_evaluator
from aodh.storage import models
from aodh.tests import constants
from aodh.tests.unit.evaluator import base


class TestEventAlarmEvaluate(base.TestEvaluatorBase):
    EVALUATOR = event_evaluator.EventAlarmEvaluator

    @staticmethod
    def _alarm(**kwargs):
        alarm_id = kwargs.get('id') or uuidutils.generate_uuid()
        return models.Alarm(name=kwargs.get('name', alarm_id),
                            type='event',
                            enabled=True,
                            alarm_id=alarm_id,
                            description='desc',
                            state=kwargs.get('state', 'insufficient data'),
                            state_reason='reason',
                            severity='critical',
                            state_timestamp=constants.MIN_DATETIME,
                            timestamp=constants.MIN_DATETIME,
                            ok_actions=[],
                            insufficient_data_actions=[],
                            alarm_actions=[],
                            repeat_actions=kwargs.get('repeat', False),
                            user_id='user',
                            project_id=kwargs.get('project', ''),
                            time_constraints=[],
                            rule=dict(event_type=kwargs.get('event_type', '*'),
                                      query=kwargs.get('query', [])))

    @staticmethod
    def _event(**kwargs):
        return {'message_id': kwargs.get('id') or uuidutils.generate_uuid(),
                'event_type': kwargs.get('event_type', 'type0'),
                'traits': kwargs.get('traits', [])}

    def _setup_alarm_storage(self, alarms):
        self._stored_alarms = {a.alarm_id: copy.deepcopy(a) for a in alarms}
        self._update_history = []

        def get_alarms(**kwargs):
            return (a for a in self._stored_alarms.values())

        def update_alarm(alarm):
            self._stored_alarms[alarm.alarm_id] = copy.deepcopy(alarm)
            self._update_history.append(dict(alarm_id=alarm.alarm_id,
                                             state=alarm.state))

        self.storage_conn.get_alarms.side_effect = get_alarms
        self.storage_conn.update_alarm.side_effect = update_alarm

    def _setup_alarm_notifier(self):
        self._notification_history = []

        def notify(alarm, previous, reason, data):
            self._notification_history.append(dict(alarm_id=alarm.alarm_id,
                                                   state=alarm.state,
                                                   previous=previous,
                                                   reason=reason,
                                                   data=data))

        self.notifier.notify.side_effect = notify

    def _do_test_event_alarm(self, alarms, events,
                             expect_db_queries=None,
                             expect_alarm_states=None,
                             expect_alarm_updates=None,
                             expect_notifications=None):
        self._setup_alarm_storage(alarms)
        self._setup_alarm_notifier()

        self.evaluator.evaluate_events(events)

        if expect_db_queries is not None:
            expected = [mock.call(enabled=True,
                                  type='event',
                                  project_id=p) for p in expect_db_queries]
            self.assertEqual(expected,
                             self.storage_conn.get_alarms.call_args_list)

        if expect_alarm_states is not None:
            for alarm_id, state in expect_alarm_states.items():
                self.assertEqual(state, self._stored_alarms[alarm_id].state)

        if expect_alarm_updates is not None:
            self.assertEqual(len(expect_alarm_updates),
                             len(self._update_history))
            for alarm, h in zip(expect_alarm_updates, self._update_history):
                expected = dict(alarm_id=alarm.alarm_id,
                                state=evaluator.ALARM)
                self.assertEqual(expected, h)

        if expect_notifications is not None:
            self.assertEqual(len(expect_notifications),
                             len(self._notification_history))
            for n, h in zip(expect_notifications, self._notification_history):
                alarm = n['alarm']
                event = n['event']
                previous = n.get('previous', evaluator.UNKNOWN)
                reason = ('Event <id=%(e)s,event_type=%(type)s> hits the '
                          'query <query=%(query)s>.') % {
                    'e': event['message_id'],
                    'type': event['event_type'],
                    'query': json.dumps(alarm.rule['query'], sort_keys=True)}
                data = {'type': 'event', 'event': event}
                expected = dict(alarm_id=alarm.alarm_id,
                                state=evaluator.ALARM,
                                previous=previous,
                                reason=reason,
                                data=data)
                self.assertEqual(expected, h)

    def test_fire_alarm_in_the_same_project_id(self):
        alarm = self._alarm(project='project1')
        event = self._event(traits=[['project_id', 1, 'project1']])
        self._do_test_event_alarm(
            [alarm], [event],
            expect_db_queries=['project1'],
            expect_alarm_states={alarm.alarm_id: evaluator.ALARM},
            expect_alarm_updates=[alarm],
            expect_notifications=[dict(alarm=alarm, event=event)])

    def test_fire_alarm_in_the_same_tenant_id(self):
        alarm = self._alarm(project='project1')
        event = self._event(traits=[['tenant_id', 1, 'project1']])
        self._do_test_event_alarm(
            [alarm], [event],
            expect_db_queries=['project1'],
            expect_alarm_states={alarm.alarm_id: evaluator.ALARM},
            expect_alarm_updates=[alarm],
            expect_notifications=[dict(alarm=alarm, event=event)])

    def test_fire_alarm_in_project_none(self):
        alarm = self._alarm(project='')
        event = self._event()
        self._do_test_event_alarm(
            [alarm], [event],
            expect_db_queries=[''],
            expect_alarm_states={alarm.alarm_id: evaluator.ALARM},
            expect_alarm_updates=[alarm],
            expect_notifications=[dict(alarm=alarm, event=event)])

    def test_continue_following_evaluation_after_exception(self):
        alarms = [
            self._alarm(id=1),
            self._alarm(id=2),
        ]
        event = self._event()

        original = self.evaluator._fire_alarm

        with mock.patch.object(event_evaluator.EventAlarmEvaluator,
                               '_fire_alarm') as _fire_alarm:
            def _side_effect(*args, **kwargs):
                _fire_alarm.side_effect = original
                return Exception('boom')

            _fire_alarm.side_effect = _side_effect

            self._do_test_event_alarm(
                alarms, [event],
                expect_alarm_states={alarms[0].alarm_id: evaluator.UNKNOWN,
                                     alarms[1].alarm_id: evaluator.ALARM},
                expect_alarm_updates=[alarms[1]],
                expect_notifications=[dict(alarm=alarms[1], event=event)])

    def test_skip_event_missing_event_type(self):
        alarm = self._alarm()
        event = {'message_id': uuidutils.generate_uuid(), 'traits': []}
        self._do_test_event_alarm(
            [alarm], [event],
            expect_alarm_states={alarm.alarm_id: evaluator.UNKNOWN},
            expect_alarm_updates=[],
            expect_notifications=[])

    def test_skip_event_missing_message_id(self):
        alarm = self._alarm()
        event = {'event_type': 'type1', 'traits': []}
        self._do_test_event_alarm(
            [alarm], [event],
            expect_alarm_states={alarm.alarm_id: evaluator.UNKNOWN},
            expect_alarm_updates=[],
            expect_notifications=[])

    def test_continue_alarming_when_repeat_actions_enabled(self):
        alarm = self._alarm(repeat=True, state=evaluator.ALARM)
        event = self._event()
        self._do_test_event_alarm(
            [alarm], [event],
            expect_alarm_states={alarm.alarm_id: evaluator.ALARM},
            expect_alarm_updates=[alarm],
            expect_notifications=[dict(alarm=alarm, event=event,
                                       previous=evaluator.ALARM)])

    def test_do_not_continue_alarming_when_repeat_actions_disabled(self):
        alarm = self._alarm(repeat=False, state=evaluator.ALARM)
        event = self._event()
        self._do_test_event_alarm(
            [alarm], [event],
            expect_alarm_states={alarm.alarm_id: evaluator.ALARM},
            expect_alarm_updates=[],
            expect_notifications=[])

    def test_skip_uninterested_event_type(self):
        alarm = self._alarm(event_type='compute.instance.exists')
        event = self._event(event_type='compute.instance.update')
        self._do_test_event_alarm(
            [alarm], [event],
            expect_alarm_states={alarm.alarm_id: evaluator.UNKNOWN},
            expect_alarm_updates=[],
            expect_notifications=[])

    def test_fire_alarm_event_type_pattern_matched(self):
        alarm = self._alarm(event_type='compute.instance.*')
        event = self._event(event_type='compute.instance.update')
        self._do_test_event_alarm(
            [alarm], [event],
            expect_alarm_states={alarm.alarm_id: evaluator.ALARM},
            expect_alarm_updates=[alarm],
            expect_notifications=[dict(alarm=alarm, event=event)])

    def test_skip_event_type_pattern_unmatched(self):
        alarm = self._alarm(event_type='compute.instance.*')
        event = self._event(event_type='dummy.compute.instance')
        self._do_test_event_alarm(
            [alarm], [event],
            expect_alarm_states={alarm.alarm_id: evaluator.UNKNOWN},
            expect_alarm_updates=[],
            expect_notifications=[])

    def test_fire_alarm_query_matched_string(self):
        alarm = self._alarm(query=[dict(field="traits.state",
                                        value="stopped",
                                        op="eq")])
        event = self._event(traits=[['state', 1, 'stopped']])
        self._do_test_event_alarm(
            [alarm], [event],
            expect_alarm_states={alarm.alarm_id: evaluator.ALARM},
            expect_alarm_updates=[alarm],
            expect_notifications=[dict(alarm=alarm, event=event)])

    def test_skip_query_unmatched_string(self):
        alarm = self._alarm(query=[dict(field="traits.state",
                                        value="stopped",
                                        op="eq")])
        event = self._event(traits=[['state', 1, 'active']])
        self._do_test_event_alarm(
            [alarm], [event],
            expect_alarm_states={alarm.alarm_id: evaluator.UNKNOWN},
            expect_alarm_updates=[],
            expect_notifications=[])

    def test_fire_alarm_query_matched_integer(self):
        alarm = self._alarm(query=[dict(field="traits.instance_type_id",
                                        type="integer",
                                        value="5",
                                        op="eq")])
        event = self._event(traits=[['instance_type_id', 2, 5]])
        self._do_test_event_alarm(
            [alarm], [event],
            expect_alarm_states={alarm.alarm_id: evaluator.ALARM},
            expect_alarm_updates=[alarm],
            expect_notifications=[dict(alarm=alarm, event=event)])

    def test_skip_query_unmatched_integer(self):
        alarm = self._alarm(query=[dict(field="traits.instance_type_id",
                                        type="integer",
                                        value="5",
                                        op="eq")])
        event = self._event(traits=[['instance_type_id', 2, 6]])
        self._do_test_event_alarm(
            [alarm], [event],
            expect_alarm_states={alarm.alarm_id: evaluator.UNKNOWN},
            expect_alarm_updates=[],
            expect_notifications=[])

    def test_fire_alarm_query_matched_float(self):
        alarm = self._alarm(query=[dict(field="traits.io_read_kbs",
                                        type="float",
                                        value="123.456",
                                        op="eq")])
        event = self._event(traits=[['io_read_kbs', 3, 123.456]])
        self._do_test_event_alarm(
            [alarm], [event],
            expect_alarm_states={alarm.alarm_id: evaluator.ALARM},
            expect_alarm_updates=[alarm],
            expect_notifications=[dict(alarm=alarm, event=event)])

    def test_skip_query_unmatched_float(self):
        alarm = self._alarm(query=[dict(field="traits.io_read_kbs",
                                        type="float",
                                        value="123.456",
                                        op="eq")])
        event = self._event(traits=[['io_read_kbs', 3, 456.123]])
        self._do_test_event_alarm(
            [alarm], [event],
            expect_alarm_states={alarm.alarm_id: evaluator.UNKNOWN},
            expect_alarm_updates=[],
            expect_notifications=[])

    def test_fire_alarm_query_matched_datetime(self):
        alarm = self._alarm(query=[dict(field="traits.created_at",
                                        type="datetime",
                                        value="2015-09-01T18:52:27.214309",
                                        op="eq")])
        event = self._event(traits=[['created_at', 4,
                                     '2015-09-01T18:52:27.214309']])
        self._do_test_event_alarm(
            [alarm], [event],
            expect_alarm_states={alarm.alarm_id: evaluator.ALARM},
            expect_alarm_updates=[alarm],
            expect_notifications=[dict(alarm=alarm, event=event)])

    def test_skip_query_unmatched_datetime(self):
        alarm = self._alarm(query=[dict(field="traits.created_at",
                                        type="datetime",
                                        value="2015-09-01T18:52:27.214309",
                                        op="eq")])
        event = self._event(traits=[['created_at', 4,
                                     '2015-09-02T18:52:27.214309']])
        self._do_test_event_alarm(
            [alarm], [event],
            expect_alarm_states={alarm.alarm_id: evaluator.UNKNOWN},
            expect_alarm_updates=[],
            expect_notifications=[])

    def test_skip_alarm_due_to_uncompareable_trait(self):
        alarm = self._alarm(query=[dict(field="traits.created_at",
                                        type="datetime",
                                        value="2015-09-01T18:52:27.214309",
                                        op="eq")])
        event = self._event(traits=[['created_at', 3, 123.456]])
        self._do_test_event_alarm(
            [alarm], [event],
            expect_alarm_states={alarm.alarm_id: evaluator.UNKNOWN},
            expect_alarm_updates=[],
            expect_notifications=[])

    def test_event_alarm_cache_hit(self):
        alarm = self._alarm(project='project2', event_type='none')
        events = [
            self._event(traits=[['project_id', 1, 'project2']]),
            self._event(traits=[['project_id', 1, 'project2']]),
        ]
        self._do_test_event_alarm([alarm], events,
                                  expect_db_queries=['project2'])

    def test_event_alarm_cache_updated_after_fired(self):
        alarm = self._alarm(project='project2', event_type='type1',
                            repeat=False)
        events = [
            self._event(event_type='type1',
                        traits=[['project_id', 1, 'project2']]),
            self._event(event_type='type1',
                        traits=[['project_id', 1, 'project2']]),
        ]
        self._do_test_event_alarm(
            [alarm], events,
            expect_db_queries=['project2'],
            expect_alarm_states={alarm.alarm_id: evaluator.ALARM},
            expect_alarm_updates=[alarm],
            expect_notifications=[dict(alarm=alarm, event=events[0])])

    def test_event_alarm_caching_disabled(self):
        alarm = self._alarm(project='project2', event_type='none')
        events = [
            self._event(traits=[['project_id', 1, 'project2']]),
            self._event(traits=[['project_id', 1, 'project2']]),
        ]
        self.evaluator.conf.event_alarm_cache_ttl = 0
        self._do_test_event_alarm([alarm], events,
                                  expect_db_queries=['project2', 'project2'])

    @mock.patch.object(timeutils, 'utcnow')
    def test_event_alarm_cache_expired(self, mock_utcnow):
        alarm = self._alarm(project='project2', event_type='none')
        events = [
            self._event(traits=[['project_id', 1, 'project2']]),
            self._event(traits=[['project_id', 1, 'project2']]),
        ]
        mock_utcnow.side_effect = [
            datetime.datetime(2015, 1, 1, 0, 0, 0),
            datetime.datetime(2015, 1, 1, 1, 0, 0),
            datetime.datetime(2015, 1, 1, 1, 1, 0),
        ]
        self._do_test_event_alarm([alarm], events,
                                  expect_db_queries=['project2', 'project2'])

    def test_event_alarm_cache_miss(self):
        events = [
            self._event(traits=[['project_id', 1, 'project2']]),
            self._event(traits=[['project_id', 1, 'project3']]),
        ]
        self._do_test_event_alarm([], events,
                                  expect_db_queries=['project2', 'project3'])



# Copyright 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""initial base

Revision ID: 12fe8fac9fe4
Revises:
Create Date: 2015-07-28 17:38:37.022899

"""

# revision identifiers, used by Alembic.
revision = '12fe8fac9fe4'
down_revision = None
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
from sqlalchemy import types

import aodh.storage.sqlalchemy.models


class PreciseTimestamp(types.TypeDecorator):
    """Represents a timestamp precise to the microsecond."""

    impl = sa.DateTime

    def load_dialect_impl(self, dialect):
        if dialect.name == 'mysql':
            return dialect.type_descriptor(
                types.DECIMAL(precision=20,
                              scale=6,
                              asdecimal=True))
        return dialect.type_descriptor(self.impl)


def upgrade():
    op.create_table(
        'alarm_history',
        sa.Column('event_id', sa.String(length=128), nullable=False),
        sa.Column('alarm_id', sa.String(length=128), nullable=True),
        sa.Column('on_behalf_of', sa.String(length=128), nullable=True),
        sa.Column('project_id', sa.String(length=128), nullable=True),
        sa.Column('user_id', sa.String(length=128), nullable=True),
        sa.Column('type', sa.String(length=20), nullable=True),
        sa.Column('detail', sa.Text(), nullable=True),
        sa.Column('timestamp',
                  PreciseTimestamp(),
                  nullable=True),
        sa.PrimaryKeyConstraint('event_id')
    )
    op.create_index(
        'ix_alarm_history_alarm_id', 'alarm_history', ['alarm_id'],
        unique=False)
    op.create_table(
        'alarm',
        sa.Column('alarm_id', sa.String(length=128), nullable=False),
        sa.Column('enabled', sa.Boolean(), nullable=True),
        sa.Column('name', sa.Text(), nullable=True),
        sa.Column('type', sa.String(length=50), nullable=True),
        sa.Column('severity', sa.String(length=50), nullable=True),
        sa.Column('description', sa.Text(), nullable=True),
        sa.Column('timestamp',
                  PreciseTimestamp(),
                  nullable=True),
        sa.Column('user_id', sa.String(length=128), nullable=True),
        sa.Column('project_id', sa.String(length=128), nullable=True),
        sa.Column('state', sa.String(length=255), nullable=True),
        sa.Column('state_timestamp',
                  PreciseTimestamp(),
                  nullable=True),
        sa.Column('ok_actions',
                  aodh.storage.sqlalchemy.models.JSONEncodedDict(),
                  nullable=True),
        sa.Column('alarm_actions',
                  aodh.storage.sqlalchemy.models.JSONEncodedDict(),
                  nullable=True),
        sa.Column('insufficient_data_actions',
                  aodh.storage.sqlalchemy.models.JSONEncodedDict(),
                  nullable=True),
        sa.Column('repeat_actions', sa.Boolean(), nullable=True),
        sa.Column('rule',
                  aodh.storage.sqlalchemy.models.JSONEncodedDict(),
                  nullable=True),
        sa.Column('time_constraints',
                  aodh.storage.sqlalchemy.models.JSONEncodedDict(),
                  nullable=True),
        sa.PrimaryKeyConstraint('alarm_id')
    )
    op.create_index(
        'ix_alarm_project_id', 'alarm', ['project_id'], unique=False)
    op.create_index(
        'ix_alarm_user_id', 'alarm', ['user_id'], unique=False)



#
# Copyright 2013-2014 eNovance
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Rest alarm notifier."""
import json

from oslo_config import cfg
from oslo_log import log
from oslo_utils import uuidutils
import requests
import urllib.parse as urlparse

from aodh import notifier

LOG = log.getLogger(__name__)

OPTS = [
    cfg.StrOpt('rest_notifier_certificate_file',
               default='',
               help='SSL Client certificate file for REST notifier.'
               ),
    cfg.StrOpt('rest_notifier_certificate_key',
               default='',
               help='SSL Client private key file for REST notifier.'
               ),
    cfg.StrOpt('rest_notifier_ca_bundle_certificate_path',
               help='SSL CA_BUNDLE certificate for REST notifier',
               ),
    cfg.BoolOpt('rest_notifier_ssl_verify',
                default=True,
                help='Whether to verify the SSL Server certificate when '
                'calling alarm action.'
                ),
    cfg.IntOpt('rest_notifier_max_retries',
               default=0,
               help='Number of retries for REST notifier',
               ),

]


class RestAlarmNotifier(notifier.AlarmNotifier):
    """Rest alarm notifier."""

    def __init__(self, conf):
        super(RestAlarmNotifier, self).__init__(conf)
        self.conf = conf

    def notify(self, action, alarm_id, alarm_name, severity, previous,
               current, reason, reason_data, headers=None):
        headers = headers or {}
        if 'x-openstack-request-id' not in headers:
            headers['x-openstack-request-id'] = b'req-' + \
                uuidutils.generate_uuid().encode('ascii')

        LOG.info(
            "Notifying alarm %(alarm_name)s %(alarm_id)s with severity"
            " %(severity)s from %(previous)s to %(current)s with action "
            "%(action)s because %(reason)s. request-id: %(request_id)s " %
            ({'alarm_name': alarm_name, 'alarm_id': alarm_id,
              'severity': severity, 'previous': previous,
              'current': current, 'action': action, 'reason': reason,
              'request_id': headers['x-openstack-request-id']}))
        body = {'alarm_name': alarm_name, 'alarm_id': alarm_id,
                'severity': severity, 'previous': previous,
                'current': current, 'reason': reason,
                'reason_data': reason_data}
        headers['content-type'] = 'application/json'
        kwargs = {'data': json.dumps(body),
                  'headers': headers}

        if action.scheme == 'https':
            default_verify = int(self.conf.rest_notifier_ssl_verify)
            options = urlparse.parse_qs(action.query)
            verify = bool(int(options.get('aodh-alarm-ssl-verify',
                                          [default_verify])[-1]))
            if verify and self.conf.rest_notifier_ca_bundle_certificate_path:
                verify = self.conf.rest_notifier_ca_bundle_certificate_path
            kwargs['verify'] = verify

            cert = self.conf.rest_notifier_certificate_file
            key = self.conf.rest_notifier_certificate_key
            if cert:
                kwargs['cert'] = (cert, key) if key else cert

        # FIXME(rhonjo): Retries are automatically done by urllib3 in requests
        # library. However, there's no interval between retries in urllib3
        # implementation. It will be better to put some interval between
        # retries (future work).
        max_retries = self.conf.rest_notifier_max_retries
        session = requests.Session()
        session.mount(action.geturl(),
                      requests.adapters.HTTPAdapter(max_retries=max_retries))
        resp = session.post(action.geturl(), **kwargs)
        LOG.info('Notifying alarm <%(id)s> gets response: %(status_code)s '
                 '%(reason)s.', {'id': alarm_id,
                                 'status_code': resp.status_code,
                                 'reason': resp.reason})
